from enum import Enum
from inspect import isawaitable
from typing import Callable, Dict, Iterable, Type

import graphene
import sqlalchemy
import sqlalchemy as sa
from graphene import Argument, Field, Interface, ObjectType, ResolveInfo
from graphene.types.mutation import MutationOptions
from graphene.types.objecttype import ObjectTypeOptions
from graphene.types.utils import yank_fields_from_attrs
from graphene.utils.get_unbound_function import get_unbound_function
from graphene.utils.props import props
from sqlalchemy.orm import DeclarativeMeta

from .get_input_type import get_input_fields, get_input_type
from .gql_fields import get_fields
from .gql_id import ResolvedGlobalId
from .types import SQLAlchemyObjectType
from .utils import get_query


class SQLMutationOptions(ObjectTypeOptions):
    model: DeclarativeMeta = None
    arguments: Dict[str, Argument] = None
    output: Type[SQLAlchemyObjectType] = None
    resolver: Callable = None
    interfaces: Iterable[Type[Interface]] = ()
    input_type: Type[graphene.InputObjectType] = None


class _BaseMutation(ObjectType):
    _meta: SQLMutationOptions

    @classmethod
    def Field(
        cls, name=None, description=None, deprecation_reason=None, required=False
    ):
        """Mount instance of mutation Field."""
        return graphene.Field(
            cls._meta.output,
            args=cls._meta.arguments,
            resolver=cls._meta.resolver,
            name=name,
            description=description or cls._meta.description,
            deprecation_reason=deprecation_reason,
            required=required,
        )

    @classmethod
    async def get_query(cls, info: ResolveInfo):
        return get_query(cls._meta.model, info, cls.__name__)

    @classmethod
    async def get_node(cls, info: ResolveInfo, id: int):
        session = info.context.session

        pk = sqlalchemy.inspect(cls._meta.model).primary_key[0]
        q = (await cls.get_query(info)).where(pk == id)
        result = cls(**(await session.execute(q)).first())
        return result


class SQLAlchemyUpdateMutation(_BaseMutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        model: Type[DeclarativeMeta],
        interfaces=(),
        resolver=None,
        output=None,
        arguments=None,
        only_fields=(),
        exclude_fields=(),
        required_fields=(),
        input_fields: dict = None,
        input_type_name: str = None,
        _meta=None,
        **options,
    ):
        if not _meta:
            _meta = SQLMutationOptions(cls)

        output = output or getattr(cls, "Output", None)
        fields = {}

        for interface in interfaces:
            assert issubclass(
                interface, Interface
            ), f'All interfaces of {cls.__name__} must be a subclass of Interface. Received "{interface}".'
            fields.update(interface._meta.fields)

        if not output:
            # If output is defined, we don't need to get the fields
            fields = {}
            for base in reversed(cls.__mro__):
                fields.update(yank_fields_from_attrs(base.__dict__, _as=Field))
            output = cls

        input_type = None
        if not arguments:
            input_class = getattr(cls, "Arguments", None)
            if input_class:
                arguments = props(input_class)
            else:
                if not input_fields:
                    input_fields = get_input_fields(
                        model,
                        only_fields=only_fields,
                        exclude_fields=exclude_fields,
                        required_fields=required_fields,
                    )
                input_type = get_input_type(
                    input_type_name or cls.__name__ + "InputType",
                    input_fields=input_fields,
                )
                arguments = {
                    "id": graphene.ID(required=True),
                    "value": graphene.Argument(input_type, required=True),
                }

        if not resolver:
            mutate = getattr(cls, "mutate", None)
            assert mutate, "All mutations must define a mutate method in it"
            resolver = get_unbound_function(mutate)

        if _meta.fields:
            _meta.fields.update(fields)
        else:
            _meta.fields = fields
        _meta.interfaces = interfaces
        _meta.output = output
        _meta.resolver = resolver
        _meta.arguments = arguments
        _meta.model = model
        _meta.input_type = input_type

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def mutate(cls, root, info: ResolveInfo, id: str, value: dict):
        session = info.context.session
        model = cls._meta.model
        output = cls._meta.output

        table = sa.inspect(model).persist_selectable
        pk = table.primary_key.columns[0]

        type_name, id_ = ResolvedGlobalId.decode(id)

        try:
            field_set = get_fields(model, info, type_name)
        except Exception as e:
            field_set = []

        if not value:
            raise Exception("No value provided")

        q = sa.update(model).values(value).where(pk == id_)

        if field_set and getattr(session.bind, "name", "") != "sqlite":
            row = (await session.execute(q.returning(*field_set))).first()
            result = output(**row)
        else:
            await session.execute(q)
            result = output.get_node(info, id_)

            if isawaitable(result):
                result = await result

        return result


class SQLAlchemyCreateMutation(_BaseMutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        model: Type[DeclarativeMeta],
        interfaces=(),
        resolver=None,
        output=None,
        arguments=None,
        only_fields=(),
        exclude_fields=(),
        required_fields=(),
        input_fields: dict = None,
        input_type_name: str = None,
        _meta=None,
        **options,
    ):
        if not _meta:
            _meta = SQLMutationOptions(cls)

        output = output or getattr(cls, "Output", None)
        fields = {}

        for interface in interfaces:
            assert issubclass(
                interface, Interface
            ), f'All interfaces of {cls.__name__} must be a subclass of Interface. Received "{interface}".'
            fields.update(interface._meta.fields)

        if not output:
            # If output is defined, we don't need to get the fields
            fields = {}
            for base in reversed(cls.__mro__):
                fields.update(yank_fields_from_attrs(base.__dict__, _as=Field))
            output = cls

        input_type = None
        if not arguments:
            input_class = getattr(cls, "Arguments", None)
            if input_class:
                arguments = props(input_class)
            else:
                if not input_fields:
                    input_fields = get_input_fields(
                        model,
                        only_fields=only_fields,
                        exclude_fields=exclude_fields,
                        required_fields=required_fields,
                    )
                input_type = get_input_type(
                    input_type_name or cls.__name__ + "InputType",
                    input_fields=input_fields,
                )
                arguments = {
                    "value": graphene.Argument(input_type, required=True),
                }

        if not resolver:
            mutate = getattr(cls, "mutate", None)
            assert mutate, "All mutations must define a mutate method in it"
            resolver = get_unbound_function(mutate)

        if _meta.fields:
            _meta.fields.update(fields)
        else:
            _meta.fields = fields
        _meta.interfaces = interfaces
        _meta.output = output
        _meta.resolver = resolver
        _meta.arguments = arguments
        _meta.model = model
        _meta.input_type = input_type

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def mutate(cls, root, info: ResolveInfo, value: dict):
        session = info.context.session
        model = cls._meta.model
        output = cls._meta.output

        try:
            field_set = get_fields(model, info, output.__name__)
        except Exception as e:
            field_set = []

        if not value:
            raise Exception("No value provided")

        q = sa.insert(model).values(
            {k: v.value if isinstance(v, Enum) else v for k, v in value.items()}
        )

        if field_set and getattr(session.bind, "name", "") != "sqlite":
            row = (await session.execute(q.returning(*field_set))).first()
            result = output(**row)
        else:
            id_ = (await session.execute(q)).inserted_primary_key[0]
            result = output.get_node(info, id_)

            if isawaitable(result):
                result = await result

        return result


class SQLAlchemyDeleteMutation(_BaseMutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        model: Type[DeclarativeMeta],
        interfaces=(),
        resolver=None,
        output=None,
        arguments=None,
        only_fields=(),
        exclude_fields=(),
        required_fields=(),
        _meta=None,
        **options,
    ):
        if not _meta:
            _meta = SQLMutationOptions(cls)

        output = output or getattr(cls, "Output", None)
        fields = {}

        for interface in interfaces:
            assert issubclass(
                interface, Interface
            ), f'All interfaces of {cls.__name__} must be a subclass of Interface. Received "{interface}".'
            fields.update(interface._meta.fields)

        if not output:
            # If output is defined, we don't need to get the fields
            fields = {}
            for base in reversed(cls.__mro__):
                fields.update(yank_fields_from_attrs(base.__dict__, _as=Field))
            output = cls

        if not arguments:
            input_class = getattr(cls, "Arguments", None)
            if input_class:
                arguments = props(input_class)
            else:
                arguments = {
                    "id": graphene.ID(required=True),
                }

        if not resolver:
            mutate = getattr(cls, "mutate", None)
            assert mutate, "All mutations must define a mutate method in it"
            resolver = get_unbound_function(mutate)

        if _meta.fields:
            _meta.fields.update(fields)
        else:
            _meta.fields = fields
        _meta.interfaces = interfaces
        _meta.output = output
        _meta.resolver = resolver
        _meta.arguments = arguments
        _meta.model = model

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def mutate(cls, root, info: ResolveInfo, id: str):
        session = info.context.session
        model = cls._meta.model
        output = cls._meta.output

        table = sa.inspect(model).persist_selectable
        pk = table.primary_key.columns[0]

        type_name, id_ = ResolvedGlobalId.decode(id)

        try:
            field_set = get_fields(model, info, type_name)
        except Exception as e:
            field_set = []

        q = sa.delete(model).where(pk == id_)

        if field_set and getattr(session.bind, "name", "") != "sqlite":
            row = (await session.execute(q.returning(*field_set))).first()
            result = output(**row)
        else:
            id_ = (await session.execute(q)).lastrowid
            result = output.get_node(info, id_)

            if isawaitable(result):
                result = await result

        return result



from typing import Callable, Optional, Type

from graphene import Dynamic, Field, List, String
from sqlalchemy import Column, ForeignKey
from sqlalchemy.orm import ColumnProperty, RelationshipProperty, interfaces

from .batching import get_batch_resolver, get_fk_resolver
from .fields import BatchSQLAlchemyConnectionField, ModelField
from .registry import Registry
from .resolvers import get_custom_resolver
from .sqlalchemy_converter import convert_sqlalchemy_type

try:
    from sqlalchemy_utils import ChoiceType, JSONType, ScalarListType, TSVectorType
except ImportError:
    ChoiceType = JSONType = ScalarListType = TSVectorType = object

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .types import SQLAlchemyObjectType


def get_column_doc(column: Column) -> Optional[str]:
    return getattr(column, "doc", None)


def is_column_nullable(column: Column) -> bool:
    return bool(getattr(column, "nullable", True))


def convert_sqlalchemy_relationship(
    relationship_prop: RelationshipProperty,
    obj_type: Type["SQLAlchemyObjectType"],
    connection_field_factory: Optional[Callable],
    orm_field_name: str,
    **field_kwargs,
) -> Dynamic:
    def dynamic_type():
        """:rtype: Field|None"""
        direction = relationship_prop.direction
        child_type = obj_type._meta.registry.get_type_for_model(
            relationship_prop.mapper.entity
        )

        if not child_type:
            return None

        if direction == interfaces.MANYTOONE or not relationship_prop.uselist:
            return _convert_o2o_or_m2o_relationship(
                relationship_prop,
                obj_type,
                orm_field_name,
                **field_kwargs,
            )

        if direction in (interfaces.ONETOMANY, interfaces.MANYTOMANY):
            return _convert_o2m_or_m2m_relationship(
                relationship_prop,
                obj_type,
                connection_field_factory,
                **field_kwargs,
            )

    return Dynamic(dynamic_type)


def convert_sqlalchemy_fk(
    fk: ForeignKey,
    obj_type: "SQLAlchemyObjectType",
    orm_field_name: str,
) -> Dynamic:
    def dynamic_type():
        child_type = obj_type._meta.registry.get_type_for_model(
            fk.constraint.referred_table
        )

        if child_type is None:
            return

        resolver = get_custom_resolver(obj_type, orm_field_name)
        if resolver is None:
            resolver = get_fk_resolver(fk, single=True)

        return Field(child_type, resolver=resolver)

    return Dynamic(dynamic_type)


def convert_sqlalchemy_fk_reverse(
    fk: ForeignKey,
    obj_type: "SQLAlchemyObjectType",
) -> Dynamic:
    def dynamic_type():
        return BatchSQLAlchemyConnectionField.from_fk(fk, obj_type._meta.registry)

    return Dynamic(dynamic_type)


def _convert_o2o_or_m2o_relationship(
    relationship_prop: RelationshipProperty,
    obj_type: Type["SQLAlchemyObjectType"],
    orm_field_name: str,
    **field_kwargs,
) -> Field:
    """
    Convert one-to-one or many-to-one relationshsip.
    Return an object field.
    """

    child_type = obj_type._meta.registry.get_type_for_model(
        relationship_prop.mapper.entity
    )

    resolver = get_custom_resolver(obj_type, orm_field_name)
    if resolver is None:
        resolver = get_batch_resolver(relationship_prop, single=True)

    return Field(child_type, resolver=resolver, **field_kwargs)


def _convert_o2m_or_m2m_relationship(
    relationship_prop: RelationshipProperty,
    obj_type: Type["SQLAlchemyObjectType"],
    connection_field_factory: Optional[Callable],
    **field_kwargs,
) -> Field:
    """
    Convert one-to-many or many-to-many relationshsip.
    Return a list field or a connection field.
    """

    child_type = obj_type._meta.registry.get_type_for_model(
        relationship_prop.mapper.entity
    )

    if not child_type._meta.connection:
        return Field(List(child_type), **field_kwargs)

    # TODO Allow override of connection_field_factory and resolver via ORMField
    if connection_field_factory is None:
        connection_field_factory = BatchSQLAlchemyConnectionField.from_relationship

    return connection_field_factory(
        relationship_prop, obj_type._meta.registry, **field_kwargs
    )


def convert_sqlalchemy_hybrid_method(hybrid_prop, resolver, **field_kwargs):
    if "type_" not in field_kwargs:
        # TODO The default type should be dependent on the type of the property propety.
        field_kwargs["type_"] = String

    return Field(resolver=resolver, **field_kwargs)


def convert_sqlalchemy_composite(composite_prop, registry, resolver):
    converter = registry.get_converter_for_composite(composite_prop.composite_class)
    if not converter:
        try:
            raise Exception(
                "Don't know how to convert the composite field %s (%s)"
                % (composite_prop, composite_prop.composite_class)
            )
        except AttributeError:
            # handle fields that are not attached to a class yet (don't have a parent)
            raise Exception(
                "Don't know how to convert the composite field %r (%s)"
                % (composite_prop, composite_prop.composite_class)
            )

    # TODO Add a way to override composite fields default parameters
    return converter(composite_prop, registry)


def _register_composite_class(cls, registry: Registry = None):
    if registry is None:
        from .registry import get_global_registry

        registry = get_global_registry()

    def inner(fn):
        registry.register_composite_converter(cls, fn)

    return inner


convert_sqlalchemy_composite.register = _register_composite_class


def convert_sqlalchemy_column(
    column_prop: ColumnProperty, registry: Registry, resolver: Callable, **field_kwargs
):
    column = column_prop.columns[0]
    field_kwargs.setdefault(
        "type_",
        convert_sqlalchemy_type(getattr(column, "type", None), column, registry),
    )
    field_kwargs.setdefault("required", not is_column_nullable(column))
    field_kwargs.setdefault("description", get_column_doc(column))

    return ModelField(resolver=resolver, model_field=column, **field_kwargs)



from base64 import b64encode
from inspect import isawaitable

from graphene import Context, ResolveInfo

from .TraceTreeBuilder import TraceTreeBuilder
from ..extension import Extension


class InlineTraceExtension(Extension):
    def __init__(self):
        self.should_trace = True
        self.tree_builder = TraceTreeBuilder()

    async def resolve(self, next_, parent, info: ResolveInfo, **kwargs):
        ftv1_header = not info.context.request.headers.get(
            "apollo-federation-include-trace", None
        )
        if ftv1_header:
            self.should_trace = False
        if parent is None and self.should_trace:
            self.tree_builder.start_timing()

        if self.should_trace and not self.tree_builder.stopped:

            end_node_trace = self.tree_builder.will_resolve_field(info)
        else:
            end_node_trace = lambda: None

        try:
            result = next_(parent, info, **kwargs)
            if isawaitable(result):
                result = await result
            return result
        finally:
            end_node_trace()

    def has_errors(self, errors, context):
        self.tree_builder.did_encounter_errors(errors, context)

    def format(self, context: Context):
        if self.should_trace:
            self.tree_builder.stop_timing()
            self.tree_builder.add_nodes_to_trace()

            ftv1 = str(b64encode(self.tree_builder.trace.SerializeToString()), "utf8")

            return {"ftv1": ftv1}



import graphene
import pytest
import sqlalchemy as sa
from graphene import Context

from alchql.fields import SQLAlchemyConnectionField
from alchql.gql_id import ResolvedGlobalId
from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.sql_mutation import (
    SQLAlchemyCreateMutation,
)
from alchql.types import SQLAlchemyObjectType
from tests import models as m
from tests.test_query import add_test_data


async def get_all_pets(session, schema):
    result = await schema.execute_async(
        """
            query {
                allPets {
                    edges {
                        node {
                            id
                            name
                        }
                    }
                }
            }
        """,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([m.Pet]),
        ],
    )
    assert not result.errors

    all_pets = [i["node"] for i in result.data["allPets"]["edges"]]

    return all_pets


@pytest.mark.asyncio
async def test_get_create_mutation(session):
    await add_test_data(session)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = m.Pet
            interfaces = (AsyncNode,)

    class MutationInsertPet(SQLAlchemyCreateMutation):
        class Meta:
            model = m.Pet
            output = PetType

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

    class Mutation(graphene.ObjectType):
        insert_pet = MutationInsertPet.Field()

    schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
    )

    all_pets = await get_all_pets(session, schema)

    assert {i["name"] for i in all_pets} == {
        "Garfield",
        "Lassie",
    }

    result = await schema.execute_async(
        """
            mutation UpdatePet($value: MutationInsertPetInputType!) {
                insertPet(value: $value) {
                    name
                }
            }
        """,
        variables={
            "value": {
                "name": "Odin",
                "petKind": "CAT",
                "hairKind": "SHORT",
            },
        },
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([m.Pet]),
        ],
    )

    assert not result.errors

    all_pets = await get_all_pets(session, schema)

    assert {i["name"] for i in all_pets} == {
        "Garfield",
        "Lassie",
        "Odin",
    }


@pytest.mark.asyncio
async def test_create_mutation_always_queries_primary_keys(session):
    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = m.Pet
            interfaces = (AsyncNode,)

    class MutationCreatePet(SQLAlchemyCreateMutation):
        class Meta:
            model = m.Pet
            output = PetType

        @classmethod
        async def mutate(cls, *args, **kwargs):
            result = await super().mutate(*args, **kwargs)
            assert result.id is not None
            return result

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

    class Mutation(graphene.ObjectType):
        insert_pet = MutationCreatePet.Field()

    schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
    )

    query = """
        mutation InsertPet($value: MutationCreatePetInputType!) {
            insertPet(value: $value) {
                name
            }
        }
    """

    result = await schema.execute_async(
        query,
        variables={
            "value": {
                "name": "asd",
                "petKind": "CAT",
                "hairKind": "SHORT",
            },
        },
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([m.Pet]),
        ],
    )

    assert not result.errors


@pytest.mark.asyncio
async def test_create_mutation_fk_with_relation(session):
    reporter_id = (await session.execute(sa.insert(m.Reporter))).lastrowid

    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = m.Reporter
            interfaces = (AsyncNode,)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = m.Pet
            interfaces = (AsyncNode,)

    class MutationCreatePet(SQLAlchemyCreateMutation):
        class Meta:
            model = m.Pet
            output = PetType

        @classmethod
        async def mutate(cls, *args, **kwargs):
            result = await super().mutate(*args, **kwargs)
            assert result.id is not None
            return result

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

    class Mutation(graphene.ObjectType):
        insert_pet = MutationCreatePet.Field()

    schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
    )

    query = """
        mutation InsertPet($value: MutationCreatePetInputType!) {
            insertPet(value: $value) {
                name
                reporterId
                reporter{
                    id
                }
            }
        }
    """

    result = await schema.execute_async(
        query,
        variables={
            "value": {
                "name": "valid",
                "petKind": "CAT",
                "hairKind": "SHORT",
                "reporterId": ResolvedGlobalId(
                    ReporterType.__name__, reporter_id
                ).encode(),
            },
        },
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([m.Pet, m.Reporter]),
        ],
    )

    assert not result.errors


@pytest.mark.asyncio
async def test_create_mutation_fk_without_relation(session, raise_graphql):
    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = m.Pet
            interfaces = (AsyncNode,)

    class MutationCreatePet(SQLAlchemyCreateMutation):
        class Meta:
            model = m.Pet
            output = PetType

        @classmethod
        async def mutate(cls, *args, **kwargs):
            result = await super().mutate(*args, **kwargs)
            assert result.id is not None
            return result

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

    class Mutation(graphene.ObjectType):
        insert_pet = MutationCreatePet.Field()

    schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
    )

    query = """
        mutation InsertPet($value: MutationCreatePetInputType!) {
            insertPet(value: $value) {
                name
                reporterId
            }
        }
    """

    result = await schema.execute_async(
        query,
        variables={
            "value": {
                "name": "asd",
                "petKind": "CAT",
                "hairKind": "SHORT",
                "reporterId": "YWRzOjQ=",
            },
        },
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([m.Pet]),
        ],
    )

    assert not result.errors


@pytest.mark.asyncio
async def test_create_mutation_rename_value(session, raise_graphql):
    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = m.Pet
            interfaces = (AsyncNode,)

    class MutationCreatePet(SQLAlchemyCreateMutation):
        class Meta:
            model = m.Pet
            output = PetType
            input_type_name = "CreatePetType"

        @classmethod
        async def mutate(cls, *args, **kwargs):
            result = await super().mutate(*args, **kwargs)
            assert result.id is not None
            return result

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

    class Mutation(graphene.ObjectType):
        insert_pet = MutationCreatePet.Field()

    schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
    )

    query = """
        mutation InsertPet($value: CreatePetType!) {
            insertPet(value: $value) {
                name
                reporterId
            }
        }
    """

    result = await schema.execute_async(
        query,
        variables={
            "value": {
                "name": "asd",
                "petKind": "CAT",
                "hairKind": "SHORT",
                "reporterId": "YWRzOjQ=",
            },
        },
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([m.Pet]),
        ],
    )

    assert not result.errors



import graphene
import pytest
import sqlalchemy as sa
from graphene import Context

from alchql.fields import SQLAlchemyConnectionField
from alchql.gql_id import ResolvedGlobalId, encode_gql_id
from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.sql_mutation import SQLAlchemyUpdateMutation
from alchql.types import SQLAlchemyObjectType
from tests.models import Pet
from tests.test_query import add_test_data


@pytest.mark.asyncio
async def test_get_update_mutation_case1(session):
    await add_test_data(session)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            interfaces = (AsyncNode,)

    class MutationUpdatePet(SQLAlchemyUpdateMutation):
        class Meta:
            model = Pet
            output = PetType

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

    class Mutation(graphene.ObjectType):
        update_pet = MutationUpdatePet.Field()

    query = """
            query {
                allPets {
                    edges {
                        node {
                            id
                            name
                        }
                    }
                }
            }
        """

    schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
    )
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )
    assert not result.errors

    all_pets = [i["node"] for i in result.data["allPets"]["edges"]]

    assert [i["name"] for i in all_pets] == [
        "Garfield",
        "Lassie",
    ]

    query = """
        mutation UpdatePet($value: MutationUpdatePetInputType!, $updatePetId: ID!) {
            updatePet(value: $value, id: $updatePetId) {
                id
                name
            }
        }
    """

    id_to_update = all_pets[0]["id"]
    new_name = "bear"

    result = await schema.execute_async(
        query,
        variables={
            "value": {"name": new_name},
            "updatePetId": id_to_update,
        },
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )

    assert not result.errors
    assert result.data["updatePet"]["id"] == id_to_update
    assert result.data["updatePet"]["name"] == new_name


@pytest.mark.asyncio
async def test_get_update_mutation_case2(session):
    await add_test_data(session)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            interfaces = (AsyncNode,)

    class MutationUpdatePet(SQLAlchemyUpdateMutation):
        class Meta:
            model = Pet

        Output = PetType

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

    class Mutation(graphene.ObjectType):
        update_pet = MutationUpdatePet.Field()

    query = """
            query {
                allPets {
                    edges {
                        node {
                            id
                            name
                        }
                    }
                }
            }
        """

    schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
    )
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )
    assert not result.errors

    all_pets = [i["node"] for i in result.data["allPets"]["edges"]]

    assert [i["name"] for i in all_pets] == [
        "Garfield",
        "Lassie",
    ]

    query = """
            mutation UpdatePet($value: MutationUpdatePetInputType!, $updatePetId: ID!) {
                updatePet(value: $value, id: $updatePetId) {
                    id
                    name
                }
            }
        """

    id_to_update = all_pets[0]["id"]
    new_name = "bear"

    result = await schema.execute_async(
        query,
        variables={
            "value": {"name": new_name},
            "updatePetId": id_to_update,
        },
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )

    assert not result.errors
    assert result.data["updatePet"]["id"] == id_to_update
    assert result.data["updatePet"]["name"] == new_name


@pytest.mark.asyncio
async def test_get_update_mutation_case3(session):
    await add_test_data(session)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            interfaces = (AsyncNode,)

    class MutationUpdatePet(SQLAlchemyUpdateMutation):
        class Meta:
            model = Pet

        id = graphene.ID()
        name = graphene.String()

        def resolve_id(self, info):
            return ResolvedGlobalId(self.__class__.__name__, self.id).encode()

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

    class Mutation(graphene.ObjectType):
        update_pet = MutationUpdatePet.Field()

    query = """
            query {
                allPets {
                    edges {
                        node {
                            id
                            name
                        }
                    }
                }
            }
        """

    schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
    )
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )
    assert not result.errors

    all_pets = [i["node"] for i in result.data["allPets"]["edges"]]

    assert [i["name"] for i in all_pets] == [
        "Garfield",
        "Lassie",
    ]

    query = """
        mutation UpdatePet($value: MutationUpdatePetInputType!, $updatePetId: ID!) {
            updatePet(value: $value, id: $updatePetId) {
                id
                name
            }
        }
    """

    id_to_update = all_pets[0]["id"]
    new_name = "bear"

    result = await schema.execute_async(
        query,
        variables={
            "value": {"name": new_name},
            "updatePetId": id_to_update,
        },
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )

    assert not result.errors
    assert result.data["updatePet"]["id"] == "TXV0YXRpb25VcGRhdGVQZXQ6MQ=="
    assert result.data["updatePet"]["name"] == new_name


@pytest.mark.asyncio
async def test_get_update_mutation_empty_value(session):
    await add_test_data(session)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            interfaces = (AsyncNode,)

    class MutationUpdatePet(SQLAlchemyUpdateMutation):
        class Meta:
            model = Pet
            output = PetType

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

    class Mutation(graphene.ObjectType):
        update_pet = MutationUpdatePet.Field()

    query = """
        query {
            allPets {
                edges {
                    node {
                        id
                        name
                    }
                }
            }
        }
    """

    schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
    )
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )
    assert not result.errors

    all_pets = [i["node"] for i in result.data["allPets"]["edges"]]

    assert [i["name"] for i in all_pets] == [
        "Garfield",
        "Lassie",
    ]

    query = """
        mutation UpdatePet($value: InputPet!, $updatePetId: ID!) {
            updatePet(value: $value, id: $updatePetId) {
                id
                name
            }
        }
    """

    id_to_update = all_pets[0]["id"]

    with pytest.raises(Exception) as e:
        result = await schema.execute_async(
            query,
            variables={
                "value": {},
                "updatePetId": id_to_update,
            },
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Pet]),
            ],
        )
        assert result.errors
        raise result.errors[0]


@pytest.mark.asyncio
async def test_update_mutation_always_queries_primary_keys(session):
    await add_test_data(session)

    global id_to_update

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            interfaces = (AsyncNode,)

    class MutationUpdatePet(SQLAlchemyUpdateMutation):
        class Meta:
            model = Pet
            output = PetType

        @classmethod
        async def mutate(cls, *args, **kwargs):
            result = await super().mutate(*args, **kwargs)
            assert result.id == id_to_update
            return result

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

    class Mutation(graphene.ObjectType):
        update_pet = MutationUpdatePet.Field()

    schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
    )

    query = """
        mutation UpdatePet($value: MutationUpdatePetInputType!, $updatePetId: ID!) {
            updatePet(value: $value, id: $updatePetId) {
                name
            }
        }
    """

    id_to_update = (await session.execute(sa.select(Pet.id))).scalars().first()
    gql_id_to_update = encode_gql_id(PetType.__name__, id_to_update)
    new_name = "New name"

    result = await schema.execute_async(
        query,
        variables={
            "value": {"name": new_name},
            "updatePetId": gql_id_to_update,
        },
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )
    assert not result.errors
    assert result.data["updatePet"]["name"] == new_name


@pytest.mark.asyncio
async def test_update_mutation_rename_value(session):
    await add_test_data(session)

    global id_to_update

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            interfaces = (AsyncNode,)

    class MutationUpdatePet(SQLAlchemyUpdateMutation):
        class Meta:
            model = Pet
            output = PetType
            input_type_name = "UpdatePetValue"

        @classmethod
        async def mutate(cls, *args, **kwargs):
            result = await super().mutate(*args, **kwargs)
            assert result.id == id_to_update
            return result

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

    class Mutation(graphene.ObjectType):
        update_pet = MutationUpdatePet.Field()

    schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
    )

    query = """
        mutation UpdatePet($value: UpdatePetValue!, $updatePetId: ID!) {
            updatePet(value: $value, id: $updatePetId) {
                name
            }
        }
    """

    id_to_update = (await session.execute(sa.select(Pet.id))).scalars().first()
    gql_id_to_update = encode_gql_id(PetType.__name__, id_to_update)
    new_name = "New name"

    result = await schema.execute_async(
        query,
        variables={
            "value": {"name": new_name},
            "updatePetId": gql_id_to_update,
        },
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )
    assert not result.errors
    assert result.data["updatePet"]["name"] == new_name



import enum
import logging
import re
from dataclasses import dataclass
from typing import List, Optional, Set

import graphene
import sqlalchemy as sa
from graphene import Dynamic, Field, Scalar
from graphql import FieldNode, ListValueNode, VariableNode
from sqlalchemy import PrimaryKeyConstraint, Table
from sqlalchemy.orm import DeclarativeMeta

from .gql_fields import camel_to_snake
from .gql_id import ResolvedGlobalId
from .utils import EnumValue, FilterItem, filter_value_to_python

RESERVED_NAMES = ["edges", "node"]
FRAGMENT = "fragment_spread"
INLINE_FRAGMENT = "inline_fragment"


@dataclass
class QueryField:
    arguments: Optional[dict]
    name: str
    values: Optional[list]


@dataclass
class FragmentField:
    name: str


class QueryHelper:
    @classmethod
    def get_filters(cls, info) -> list:
        object_types = getattr(info.context, "object_types", {})
        object_type = object_types.get(info.field_name)

        if not object_type:
            return []

        if not hasattr(object_type, "parsed_filters"):
            return []

        filters_to_apply = []

        parsed_filters = object_type.parsed_filters

        gql_field = cls.get_current_field(info)

        if gql_field and gql_field.arguments:
            for name, value in gql_field.arguments.items():
                if name not in parsed_filters:
                    continue

                if value is None:
                    continue

                filter_item: FilterItem
                filter_item = parsed_filters[name]
                if not filter_item.filter_func:
                    continue

                if hasattr(filter_item.field_type, "parse_value"):
                    value = filter_item.field_type.parse_value(value)

                if filter_item.field_type == graphene.ID:
                    global_id = ResolvedGlobalId.decode(value)
                    value = global_id.id

                if (
                    filter_item.field_type == graphene.List(of_type=graphene.ID)
                    and filter_item.field_type.of_type == graphene.ID
                ):
                    new_value = []
                    for item in value:
                        global_id = ResolvedGlobalId.decode(item)
                        new_value.append(global_id.id)
                    value = new_value

                value = parsed_filters[name].value_func(value)
                field_expr = parsed_filters[name].filter_func(value)
                filters_to_apply.append(field_expr)
        return filters_to_apply

    @classmethod
    def parse_query(cls, info) -> List[QueryField]:
        # TODO: cache was removed because of bug with two diff. fragments
        #  with the same ObjectType and different fields in each fragment
        # if not hasattr(
        #     info.context, "parsed_query"
        # ) or not info.context.parsed_query.get(info.field_name):

        nodes = info.field_nodes

        variables = info.variable_values
        object_types = getattr(info.context, "object_types", {})
        object_type = object_types.get(info.field_name)
        object_type_name = object_type.__name__ if object_type else None
        result = cls.__parse_nodes(nodes, variables, object_type_name)
        fragments = cls.__parse_fragments(info.fragments, variables)

        result = cls.__set_fragment_fields(result, fragments)
        return result
        # setattr(info.context, "parsed_query", {info.field_name: result})

        # return info.context.parsed_query[info.field_name].copy()

    @classmethod
    def get_selected_fields(cls, info, model, sort=None):
        object_types = getattr(info.context, "object_types", {})
        object_type = object_types.get(info.field_name)

        if not object_type:
            return

        gql_field = cls.get_current_field(info)

        object_type_fields = {}

        for _name in dir(object_type):
            if _name.startswith("_"):
                continue

            attr = getattr(object_type, _name, None)
            if attr and isinstance(attr, (Field, Scalar)):
                if hasattr(attr, "kwargs") and attr.kwargs.get("name"):
                    object_type_fields[attr.kwargs.get("name")] = attr
                else:
                    object_type_fields[_name] = attr

        type_ = info.context.object_types[info.field_name]
        meta_fields = type_._meta.fields

        select_fields = set()
        if isinstance(model, Table):
            for constraint in model.constraints:
                if isinstance(constraint, PrimaryKeyConstraint):
                    for i in constraint.columns:
                        select_fields.add(i)
        elif isinstance(model, DeclarativeMeta):
            select_fields.add(sa.inspect(model).primary_key[0])

        field_names_to_process = {f.name for f in gql_field.values}

        sort_field_names = set()
        if sort is not None:
            if not isinstance(sort, list):
                sort = [sort]
            for item in sort:
                if isinstance(item, (EnumValue, enum.Enum)):
                    field_name = "_".join(item.name.lower().split("_")[:-1])
                    sort_field_names.add(field_name)
                else:
                    sort_field_names.add(item)

        field_names_to_process.update(sort_field_names)
        for field in field_names_to_process:
            current_field = object_type_fields.get(field) or meta_fields.get(field)

            if current_field is None:
                continue

            if isinstance(current_field, Dynamic) and isinstance(
                current_field.type(), Field
            ):
                model_field = getattr(object_type._meta.model, field, None)
                if model_field is not None:
                    columns = model_field.prop.local_columns
                    relation_key = next(iter(columns))
                    select_fields.add(relation_key)
                else:
                    mapped_table = (
                        model
                        if isinstance(model, Table)
                        else sa.inspect(model).persist_selectable
                    )

                    for fk in mapped_table.foreign_keys:
                        if re.sub(r"_(?:id|pk)$", "", fk.parent.key) == field:
                            select_fields.add(fk.parent)
                            break
                    else:
                        logging.warning(
                            f"No field {field!r} in {object_type._meta.model.__name__}"
                        )

            model_field = getattr(current_field, "model_field", None)
            if model_field is not None:
                # labled columns could create name conflict
                if (
                    getattr(current_field, "use_label", True)
                    and field != model_field.key
                ):
                    select_fields.add(model_field.label(field))
                else:
                    select_fields.add(model_field)

        return select_fields

    @classmethod
    def get_current_field(cls, info) -> Optional[QueryField]:
        gql_fields = cls.parse_query(info)
        field_name = camel_to_snake(info.field_name)

        while gql_fields:
            next_field = gql_fields.pop()
            if next_field.name == field_name:
                return next_field

            if next_field.values:
                gql_fields.extend(next_field.values)

    @classmethod
    def has_arg(cls, info, arg: str):
        current_field = cls.get_current_field(info)

        if current_field and current_field.arguments:
            for name, value in current_field.arguments.items():
                if name == arg and value is not None:
                    return True

        return False

    @classmethod
    def get_page_info_fields(cls, info) -> Set[str]:
        current_field = cls.get_current_field(info)

        if current_field and current_field.values:
            for internal_field in current_field.values:
                if internal_field.name == "page_info":
                    return {_field.name for _field in internal_field.values}

        return set()

    @classmethod
    def __parse_nodes(cls, nodes, variables, object_type_name=None) -> list:
        values = []
        node: FieldNode
        for node in nodes:
            if node.kind == INLINE_FRAGMENT:
                if node.type_condition.name.value == object_type_name:
                    node_values = cls.__parse_nodes(
                        node.selection_set.selections, variables, object_type_name
                    )
                    return node_values
                continue

            name = camel_to_snake(node.name.value)
            node_values = None
            if node.kind == FRAGMENT:
                values.append(FragmentField(name=name))
                continue

            if node.selection_set:
                node_values = cls.__parse_nodes(
                    node.selection_set.selections, variables, object_type_name
                )

            arguments = {}
            if node.arguments:
                for arg in node.arguments:
                    if isinstance(arg.value, ListValueNode):
                        value = []
                        for arg_value in arg.value.values:
                            value.append(arg_value.value)

                    elif isinstance(arg.value, VariableNode):
                        value = variables.get(arg.value.name.value)
                    else:
                        value = arg.value.value

                    arguments[camel_to_snake(arg.name.value)] = filter_value_to_python(
                        value
                    )

            if name in RESERVED_NAMES:
                values.extend(node_values)
            else:
                values.append(
                    QueryField(name=name, values=node_values, arguments=arguments)
                )

        return values

    @classmethod
    def __parse_fragments(cls, fragments: dict, variables: dict) -> dict:
        result = {}
        for name, fragment in fragments.items():
            result[camel_to_snake(name)] = cls.__parse_nodes(
                fragment.selection_set.selections, variables
            )

        return result

    @classmethod
    def __set_fragment_fields(cls, parsed_query, fragments) -> list:
        new_values = []
        for field in parsed_query:
            if isinstance(field, FragmentField):
                extra_fields = fragments.get(field.name)
                for extra_field in extra_fields:
                    if extra_field.values:
                        extra_field.values = cls.__set_fragment_fields(
                            parsed_query=extra_field.values, fragments=fragments
                        )
                new_values.extend(extra_fields)
            else:
                if field.values:
                    field.values = cls.__set_fragment_fields(
                        parsed_query=field.values, fragments=fragments
                    )

                new_values.append(field)
        return new_values

    @classmethod
    def __get_query_field_page_info_fields(cls, field) -> Set[str]:
        if field.values:
            for internal_field in field.values:
                if internal_field.name == "page_info":
                    return {_field.name for _field in internal_field.values}

        return set()



from typing import Type

import graphene
from graphene.types import scalars
from sqlalchemy.orm import InstrumentedAttribute


def _construct_type(type_: Type):
    def init_(self, *args, model_field=None, **kwargs):
        super(type_, self).__init__(*args, **kwargs)
        # fix for unique values
        if isinstance(model_field, InstrumentedAttribute):
            model_field = model_field.prop.expression
        self.model_field = model_field

    return type(type_.__name__, (type_,), {"__init__": init_})


ID = _construct_type(scalars.ID)
Int = _construct_type(scalars.Int)
BigInt = _construct_type(scalars.BigInt)
Float = _construct_type(scalars.Float)
String = _construct_type(scalars.String)
Boolean = _construct_type(scalars.Boolean)

List = _construct_type(graphene.List)
NonNull = _construct_type(graphene.NonNull)
Date = _construct_type(graphene.Date)



import json

from graphql import GraphQLError

from .generated import reports_pb2 as Trace


def hr_timestamp_to_nanos(timestamp):
    return (timestamp.seconds * (10**9)) + timestamp.nanos


def response_path_as_string(p=None):
    if p is None:
        return ""

    res = p.key
    p = p.prev
    while p is not None:
        res = str(p.key) + "." + str(res)
        p = p.prev
    return res


def create_location_message(loc):
    location = Trace.Trace().Location()
    location.line = loc[0]
    location.column = loc[1]
    return location


def encode_graphql_error(error):
    if isinstance(error, GraphQLError):
        return {
            "message": error.message,
            "nodes": error.nodes,
            "source": error.source,
            "positions": error.positions,
            "locations": error.locations,
            "path": error.path,
            "original_error": error.original_error,
            "extensions": error.extensions,
        }


def error_to_protobuf_error(error):
    trace_error = Trace.Trace().Error()
    trace_error.message = error.message
    if len(error.locations):
        for loc in error.locations:
            location_message = create_location_message(loc)
            trace_error.location.append(location_message)
    trace_error.json = json.dumps(error, default=encode_graphql_error)
    return trace_error



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: reports.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\rreports.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\xa5\x14\n\x05Trace\x12.\n\nstart_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x13\n\x0b\x64uration_ns\x18\x0b \x01(\x04\x12\x19\n\x04root\x18\x0e \x01(\x0b\x32\x0b.Trace.Node\x12\x11\n\tsignature\x18\x13 \x01(\t\x12\x1f\n\x17unexecutedOperationBody\x18\x1b \x01(\t\x12\x1f\n\x17unexecutedOperationName\x18\x1c \x01(\t\x12\x1f\n\x07\x64\x65tails\x18\x06 \x01(\x0b\x32\x0e.Trace.Details\x12\x13\n\x0b\x63lient_name\x18\x07 \x01(\t\x12\x16\n\x0e\x63lient_version\x18\x08 \x01(\t\x12\x19\n\x04http\x18\n \x01(\x0b\x32\x0b.Trace.HTTP\x12(\n\x0c\x63\x61\x63he_policy\x18\x12 \x01(\x0b\x32\x12.Trace.CachePolicy\x12(\n\nquery_plan\x18\x1a \x01(\x0b\x32\x14.Trace.QueryPlanNode\x12\x1c\n\x14\x66ull_query_cache_hit\x18\x14 \x01(\x08\x12\x1b\n\x13persisted_query_hit\x18\x15 \x01(\x08\x12 \n\x18persisted_query_register\x18\x16 \x01(\x08\x12\x1c\n\x14registered_operation\x18\x18 \x01(\x08\x12\x1b\n\x13\x66orbidden_operation\x18\x19 \x01(\x08\x12\x1e\n\x16\x66ield_execution_weight\x18\x1f \x01(\x01\x1ay\n\x0b\x43\x61\x63hePolicy\x12\'\n\x05scope\x18\x01 \x01(\x0e\x32\x18.Trace.CachePolicy.Scope\x12\x12\n\nmax_age_ns\x18\x02 \x01(\x03"-\n\x05Scope\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06PUBLIC\x10\x01\x12\x0b\n\x07PRIVATE\x10\x02\x1a\x92\x01\n\x07\x44\x65tails\x12\x39\n\x0evariables_json\x18\x04 \x03(\x0b\x32!.Trace.Details.VariablesJsonEntry\x12\x16\n\x0eoperation_name\x18\x03 \x01(\t\x1a\x34\n\x12VariablesJsonEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1aZ\n\x05\x45rror\x12\x0f\n\x07message\x18\x01 \x01(\t\x12!\n\x08location\x18\x02 \x03(\x0b\x32\x0f.Trace.Location\x12\x0f\n\x07time_ns\x18\x03 \x01(\x04\x12\x0c\n\x04json\x18\x04 \x01(\t\x1a\x9c\x04\n\x04HTTP\x12"\n\x06method\x18\x01 \x01(\x0e\x32\x12.Trace.HTTP.Method\x12\x0c\n\x04host\x18\x02 \x01(\t\x12\x0c\n\x04path\x18\x03 \x01(\t\x12\x38\n\x0frequest_headers\x18\x04 \x03(\x0b\x32\x1f.Trace.HTTP.RequestHeadersEntry\x12:\n\x10response_headers\x18\x05 \x03(\x0b\x32 .Trace.HTTP.ResponseHeadersEntry\x12\x13\n\x0bstatus_code\x18\x06 \x01(\r\x12\x0e\n\x06secure\x18\x08 \x01(\x08\x12\x10\n\x08protocol\x18\t \x01(\t\x1a\x17\n\x06Values\x12\r\n\x05value\x18\x01 \x03(\t\x1aI\n\x13RequestHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.Trace.HTTP.Values:\x02\x38\x01\x1aJ\n\x14ResponseHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.Trace.HTTP.Values:\x02\x38\x01"w\n\x06Method\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07OPTIONS\x10\x01\x12\x07\n\x03GET\x10\x02\x12\x08\n\x04HEAD\x10\x03\x12\x08\n\x04POST\x10\x04\x12\x07\n\x03PUT\x10\x05\x12\n\n\x06\x44\x45LETE\x10\x06\x12\t\n\x05TRACE\x10\x07\x12\x0b\n\x07\x43ONNECT\x10\x08\x12\t\n\x05PATCH\x10\t\x1a(\n\x08Location\x12\x0c\n\x04line\x18\x01 \x01(\r\x12\x0e\n\x06\x63olumn\x18\x02 \x01(\r\x1a\x85\x02\n\x04Node\x12\x17\n\rresponse_name\x18\x01 \x01(\tH\x00\x12\x0f\n\x05index\x18\x02 \x01(\rH\x00\x12\x1b\n\x13original_field_name\x18\x0e \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\x13\n\x0bparent_type\x18\r \x01(\t\x12(\n\x0c\x63\x61\x63he_policy\x18\x05 \x01(\x0b\x32\x12.Trace.CachePolicy\x12\x12\n\nstart_time\x18\x08 \x01(\x04\x12\x10\n\x08\x65nd_time\x18\t \x01(\x04\x12\x1b\n\x05\x65rror\x18\x0b \x03(\x0b\x32\x0c.Trace.Error\x12\x1a\n\x05\x63hild\x18\x0c \x03(\x0b\x32\x0b.Trace.NodeB\x04\n\x02idJ\x04\x08\x04\x10\x05\x1a\xe2\x05\n\rQueryPlanNode\x12\x35\n\x08sequence\x18\x01 \x01(\x0b\x32!.Trace.QueryPlanNode.SequenceNodeH\x00\x12\x35\n\x08parallel\x18\x02 \x01(\x0b\x32!.Trace.QueryPlanNode.ParallelNodeH\x00\x12/\n\x05\x66\x65tch\x18\x03 \x01(\x0b\x32\x1e.Trace.QueryPlanNode.FetchNodeH\x00\x12\x33\n\x07\x66latten\x18\x04 \x01(\x0b\x32 .Trace.QueryPlanNode.FlattenNodeH\x00\x1a\x33\n\x0cSequenceNode\x12#\n\x05nodes\x18\x01 \x03(\x0b\x32\x14.Trace.QueryPlanNode\x1a\x33\n\x0cParallelNode\x12#\n\x05nodes\x18\x01 \x03(\x0b\x32\x14.Trace.QueryPlanNode\x1a\xd2\x01\n\tFetchNode\x12\x14\n\x0cservice_name\x18\x01 \x01(\t\x12\x1c\n\x14trace_parsing_failed\x18\x02 \x01(\x08\x12\x15\n\x05trace\x18\x03 \x01(\x0b\x32\x06.Trace\x12\x18\n\x10sent_time_offset\x18\x04 \x01(\x04\x12-\n\tsent_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\rreceived_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1ar\n\x0b\x46lattenNode\x12?\n\rresponse_path\x18\x01 \x03(\x0b\x32(.Trace.QueryPlanNode.ResponsePathElement\x12"\n\x04node\x18\x02 \x01(\x0b\x32\x14.Trace.QueryPlanNode\x1a\x42\n\x13ResponsePathElement\x12\x14\n\nfield_name\x18\x01 \x01(\tH\x00\x12\x0f\n\x05index\x18\x02 \x01(\rH\x00\x42\x04\n\x02idB\x06\n\x04nodeJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\t\x10\nJ\x04\x08\x0c\x10\rJ\x04\x08\r\x10\x0eJ\x04\x08\x17\x10\x18"\xaf\x01\n\x0cReportHeader\x12\x11\n\tgraph_ref\x18\x0c \x01(\t\x12\x10\n\x08hostname\x18\x05 \x01(\t\x12\x15\n\ragent_version\x18\x06 \x01(\t\x12\x17\n\x0fservice_version\x18\x07 \x01(\t\x12\x17\n\x0fruntime_version\x18\x08 \x01(\t\x12\r\n\x05uname\x18\t \x01(\t\x12\x1c\n\x14\x65xecutable_schema_id\x18\x0b \x01(\tJ\x04\x08\x03\x10\x04"\xbd\x01\n\x0ePathErrorStats\x12/\n\x08\x63hildren\x18\x01 \x03(\x0b\x32\x1d.PathErrorStats.ChildrenEntry\x12\x14\n\x0c\x65rrors_count\x18\x04 \x01(\x04\x12"\n\x1arequests_with_errors_count\x18\x05 \x01(\x04\x1a@\n\rChildrenEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1e\n\x05value\x18\x02 \x01(\x0b\x32\x0f.PathErrorStats:\x02\x38\x01"\xcf\x03\n\x11QueryLatencyStats\x12\x15\n\rlatency_count\x18\r \x03(\x12\x12\x15\n\rrequest_count\x18\x02 \x01(\x04\x12\x12\n\ncache_hits\x18\x03 \x01(\x04\x12\x1c\n\x14persisted_query_hits\x18\x04 \x01(\x04\x12\x1e\n\x16persisted_query_misses\x18\x05 \x01(\x04\x12\x1b\n\x13\x63\x61\x63he_latency_count\x18\x0e \x03(\x12\x12)\n\x10root_error_stats\x18\x07 \x01(\x0b\x32\x0f.PathErrorStats\x12"\n\x1arequests_with_errors_count\x18\x08 \x01(\x04\x12\x1e\n\x16public_cache_ttl_count\x18\x0f \x03(\x12\x12\x1f\n\x17private_cache_ttl_count\x18\x10 \x03(\x12\x12"\n\x1aregistered_operation_count\x18\x0b \x01(\x04\x12!\n\x19\x66orbidden_operation_count\x18\x0c \x01(\x04\x12.\n&requests_without_field_instrumentation\x18\x11 \x01(\x04J\x04\x08\x01\x10\x02J\x04\x08\x06\x10\x07J\x04\x08\t\x10\nJ\x04\x08\n\x10\x0b"A\n\x0cStatsContext\x12\x13\n\x0b\x63lient_name\x18\x02 \x01(\t\x12\x16\n\x0e\x63lient_version\x18\x03 \x01(\tJ\x04\x08\x01\x10\x02"r\n\x1f\x43ontextualizedQueryLatencyStats\x12/\n\x13query_latency_stats\x18\x01 \x01(\x0b\x32\x12.QueryLatencyStats\x12\x1e\n\x07\x63ontext\x18\x02 \x01(\x0b\x32\r.StatsContext"\xba\x01\n\x17\x43ontextualizedTypeStats\x12\x1e\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\r.StatsContext\x12@\n\rper_type_stat\x18\x02 \x03(\x0b\x32).ContextualizedTypeStats.PerTypeStatEntry\x1a=\n\x10PerTypeStatEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x18\n\x05value\x18\x02 \x01(\x0b\x32\t.TypeStat:\x02\x38\x01"\xce\x01\n\tFieldStat\x12\x13\n\x0breturn_type\x18\x03 \x01(\t\x12\x14\n\x0c\x65rrors_count\x18\x04 \x01(\x04\x12 \n\x18observed_execution_count\x18\x05 \x01(\x04\x12!\n\x19\x65stimated_execution_count\x18\n \x01(\x04\x12"\n\x1arequests_with_errors_count\x18\x06 \x01(\x04\x12\x15\n\rlatency_count\x18\t \x03(\x12J\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\x07\x10\x08J\x04\x08\x08\x10\t"\x8c\x01\n\x08TypeStat\x12\x33\n\x0eper_field_stat\x18\x03 \x03(\x0b\x32\x1b.TypeStat.PerFieldStatEntry\x1a?\n\x11PerFieldStatEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x19\n\x05value\x18\x02 \x01(\x0b\x32\n.FieldStat:\x02\x38\x01J\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03"D\n\x17ReferencedFieldsForType\x12\x13\n\x0b\x66ield_names\x18\x01 \x03(\t\x12\x14\n\x0cis_interface\x18\x02 \x01(\x08"\xd4\x01\n\x06Report\x12\x1d\n\x06header\x18\x01 \x01(\x0b\x32\r.ReportHeader\x12\x35\n\x10traces_per_query\x18\x05 \x03(\x0b\x32\x1b.Report.TracesPerQueryEntry\x12,\n\x08\x65nd_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a\x46\n\x13TracesPerQueryEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1e\n\x05value\x18\x02 \x01(\x0b\x32\x0f.TracesAndStats:\x02\x38\x01"\xe3\x01\n\x13\x43ontextualizedStats\x12\x1e\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\r.StatsContext\x12/\n\x13query_latency_stats\x18\x02 \x01(\x0b\x32\x12.QueryLatencyStats\x12<\n\rper_type_stat\x18\x03 \x03(\x0b\x32%.ContextualizedStats.PerTypeStatEntry\x1a=\n\x10PerTypeStatEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x18\n\x05value\x18\x02 \x01(\x0b\x32\t.TypeStat:\x02\x38\x01"\xb9\x02\n\x0eTracesAndStats\x12\x15\n\x05trace\x18\x01 \x03(\x0b\x32\x06.Trace\x12\x30\n\x12stats_with_context\x18\x02 \x03(\x0b\x32\x14.ContextualizedStats\x12N\n\x19referenced_fields_by_type\x18\x04 \x03(\x0b\x32+.TracesAndStats.ReferencedFieldsByTypeEntry\x12\x35\n%internal_traces_contributing_to_stats\x18\x03 \x03(\x0b\x32\x06.Trace\x1aW\n\x1bReferencedFieldsByTypeEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\'\n\x05value\x18\x02 \x01(\x0b\x32\x18.ReferencedFieldsForType:\x02\x38\x01\x62\x06proto3'
)


_TRACE = DESCRIPTOR.message_types_by_name["Trace"]
_TRACE_CACHEPOLICY = _TRACE.nested_types_by_name["CachePolicy"]
_TRACE_DETAILS = _TRACE.nested_types_by_name["Details"]
_TRACE_DETAILS_VARIABLESJSONENTRY = _TRACE_DETAILS.nested_types_by_name[
    "VariablesJsonEntry"
]
_TRACE_ERROR = _TRACE.nested_types_by_name["Error"]
_TRACE_HTTP = _TRACE.nested_types_by_name["HTTP"]
_TRACE_HTTP_VALUES = _TRACE_HTTP.nested_types_by_name["Values"]
_TRACE_HTTP_REQUESTHEADERSENTRY = _TRACE_HTTP.nested_types_by_name[
    "RequestHeadersEntry"
]
_TRACE_HTTP_RESPONSEHEADERSENTRY = _TRACE_HTTP.nested_types_by_name[
    "ResponseHeadersEntry"
]
_TRACE_LOCATION = _TRACE.nested_types_by_name["Location"]
_TRACE_NODE = _TRACE.nested_types_by_name["Node"]
_TRACE_QUERYPLANNODE = _TRACE.nested_types_by_name["QueryPlanNode"]
_TRACE_QUERYPLANNODE_SEQUENCENODE = _TRACE_QUERYPLANNODE.nested_types_by_name[
    "SequenceNode"
]
_TRACE_QUERYPLANNODE_PARALLELNODE = _TRACE_QUERYPLANNODE.nested_types_by_name[
    "ParallelNode"
]
_TRACE_QUERYPLANNODE_FETCHNODE = _TRACE_QUERYPLANNODE.nested_types_by_name["FetchNode"]
_TRACE_QUERYPLANNODE_FLATTENNODE = _TRACE_QUERYPLANNODE.nested_types_by_name[
    "FlattenNode"
]
_TRACE_QUERYPLANNODE_RESPONSEPATHELEMENT = _TRACE_QUERYPLANNODE.nested_types_by_name[
    "ResponsePathElement"
]
_REPORTHEADER = DESCRIPTOR.message_types_by_name["ReportHeader"]
_PATHERRORSTATS = DESCRIPTOR.message_types_by_name["PathErrorStats"]
_PATHERRORSTATS_CHILDRENENTRY = _PATHERRORSTATS.nested_types_by_name["ChildrenEntry"]
_QUERYLATENCYSTATS = DESCRIPTOR.message_types_by_name["QueryLatencyStats"]
_STATSCONTEXT = DESCRIPTOR.message_types_by_name["StatsContext"]
_CONTEXTUALIZEDQUERYLATENCYSTATS = DESCRIPTOR.message_types_by_name[
    "ContextualizedQueryLatencyStats"
]
_CONTEXTUALIZEDTYPESTATS = DESCRIPTOR.message_types_by_name["ContextualizedTypeStats"]
_CONTEXTUALIZEDTYPESTATS_PERTYPESTATENTRY = (
    _CONTEXTUALIZEDTYPESTATS.nested_types_by_name["PerTypeStatEntry"]
)
_FIELDSTAT = DESCRIPTOR.message_types_by_name["FieldStat"]
_TYPESTAT = DESCRIPTOR.message_types_by_name["TypeStat"]
_TYPESTAT_PERFIELDSTATENTRY = _TYPESTAT.nested_types_by_name["PerFieldStatEntry"]
_REFERENCEDFIELDSFORTYPE = DESCRIPTOR.message_types_by_name["ReferencedFieldsForType"]
_REPORT = DESCRIPTOR.message_types_by_name["Report"]
_REPORT_TRACESPERQUERYENTRY = _REPORT.nested_types_by_name["TracesPerQueryEntry"]
_CONTEXTUALIZEDSTATS = DESCRIPTOR.message_types_by_name["ContextualizedStats"]
_CONTEXTUALIZEDSTATS_PERTYPESTATENTRY = _CONTEXTUALIZEDSTATS.nested_types_by_name[
    "PerTypeStatEntry"
]
_TRACESANDSTATS = DESCRIPTOR.message_types_by_name["TracesAndStats"]
_TRACESANDSTATS_REFERENCEDFIELDSBYTYPEENTRY = _TRACESANDSTATS.nested_types_by_name[
    "ReferencedFieldsByTypeEntry"
]
_TRACE_CACHEPOLICY_SCOPE = _TRACE_CACHEPOLICY.enum_types_by_name["Scope"]
_TRACE_HTTP_METHOD = _TRACE_HTTP.enum_types_by_name["Method"]
Trace = _reflection.GeneratedProtocolMessageType(
    "Trace",
    (_message.Message,),
    {
        "CachePolicy": _reflection.GeneratedProtocolMessageType(
            "CachePolicy",
            (_message.Message,),
            {
                "DESCRIPTOR": _TRACE_CACHEPOLICY,
                "__module__": "reports_pb2"
                # @@protoc_insertion_point(class_scope:Trace.CachePolicy)
            },
        ),
        "Details": _reflection.GeneratedProtocolMessageType(
            "Details",
            (_message.Message,),
            {
                "VariablesJsonEntry": _reflection.GeneratedProtocolMessageType(
                    "VariablesJsonEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _TRACE_DETAILS_VARIABLESJSONENTRY,
                        "__module__": "reports_pb2"
                        # @@protoc_insertion_point(class_scope:Trace.Details.VariablesJsonEntry)
                    },
                ),
                "DESCRIPTOR": _TRACE_DETAILS,
                "__module__": "reports_pb2"
                # @@protoc_insertion_point(class_scope:Trace.Details)
            },
        ),
        "Error": _reflection.GeneratedProtocolMessageType(
            "Error",
            (_message.Message,),
            {
                "DESCRIPTOR": _TRACE_ERROR,
                "__module__": "reports_pb2"
                # @@protoc_insertion_point(class_scope:Trace.Error)
            },
        ),
        "HTTP": _reflection.GeneratedProtocolMessageType(
            "HTTP",
            (_message.Message,),
            {
                "Values": _reflection.GeneratedProtocolMessageType(
                    "Values",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _TRACE_HTTP_VALUES,
                        "__module__": "reports_pb2"
                        # @@protoc_insertion_point(class_scope:Trace.HTTP.Values)
                    },
                ),
                "RequestHeadersEntry": _reflection.GeneratedProtocolMessageType(
                    "RequestHeadersEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _TRACE_HTTP_REQUESTHEADERSENTRY,
                        "__module__": "reports_pb2"
                        # @@protoc_insertion_point(class_scope:Trace.HTTP.RequestHeadersEntry)
                    },
                ),
                "ResponseHeadersEntry": _reflection.GeneratedProtocolMessageType(
                    "ResponseHeadersEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _TRACE_HTTP_RESPONSEHEADERSENTRY,
                        "__module__": "reports_pb2"
                        # @@protoc_insertion_point(class_scope:Trace.HTTP.ResponseHeadersEntry)
                    },
                ),
                "DESCRIPTOR": _TRACE_HTTP,
                "__module__": "reports_pb2"
                # @@protoc_insertion_point(class_scope:Trace.HTTP)
            },
        ),
        "Location": _reflection.GeneratedProtocolMessageType(
            "Location",
            (_message.Message,),
            {
                "DESCRIPTOR": _TRACE_LOCATION,
                "__module__": "reports_pb2"
                # @@protoc_insertion_point(class_scope:Trace.Location)
            },
        ),
        "Node": _reflection.GeneratedProtocolMessageType(
            "Node",
            (_message.Message,),
            {
                "DESCRIPTOR": _TRACE_NODE,
                "__module__": "reports_pb2"
                # @@protoc_insertion_point(class_scope:Trace.Node)
            },
        ),
        "QueryPlanNode": _reflection.GeneratedProtocolMessageType(
            "QueryPlanNode",
            (_message.Message,),
            {
                "SequenceNode": _reflection.GeneratedProtocolMessageType(
                    "SequenceNode",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _TRACE_QUERYPLANNODE_SEQUENCENODE,
                        "__module__": "reports_pb2"
                        # @@protoc_insertion_point(class_scope:Trace.QueryPlanNode.SequenceNode)
                    },
                ),
                "ParallelNode": _reflection.GeneratedProtocolMessageType(
                    "ParallelNode",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _TRACE_QUERYPLANNODE_PARALLELNODE,
                        "__module__": "reports_pb2"
                        # @@protoc_insertion_point(class_scope:Trace.QueryPlanNode.ParallelNode)
                    },
                ),
                "FetchNode": _reflection.GeneratedProtocolMessageType(
                    "FetchNode",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _TRACE_QUERYPLANNODE_FETCHNODE,
                        "__module__": "reports_pb2"
                        # @@protoc_insertion_point(class_scope:Trace.QueryPlanNode.FetchNode)
                    },
                ),
                "FlattenNode": _reflection.GeneratedProtocolMessageType(
                    "FlattenNode",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _TRACE_QUERYPLANNODE_FLATTENNODE,
                        "__module__": "reports_pb2"
                        # @@protoc_insertion_point(class_scope:Trace.QueryPlanNode.FlattenNode)
                    },
                ),
                "ResponsePathElement": _reflection.GeneratedProtocolMessageType(
                    "ResponsePathElement",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _TRACE_QUERYPLANNODE_RESPONSEPATHELEMENT,
                        "__module__": "reports_pb2"
                        # @@protoc_insertion_point(class_scope:Trace.QueryPlanNode.ResponsePathElement)
                    },
                ),
                "DESCRIPTOR": _TRACE_QUERYPLANNODE,
                "__module__": "reports_pb2"
                # @@protoc_insertion_point(class_scope:Trace.QueryPlanNode)
            },
        ),
        "DESCRIPTOR": _TRACE,
        "__module__": "reports_pb2"
        # @@protoc_insertion_point(class_scope:Trace)
    },
)
_sym_db.RegisterMessage(Trace)
_sym_db.RegisterMessage(Trace.CachePolicy)
_sym_db.RegisterMessage(Trace.Details)
_sym_db.RegisterMessage(Trace.Details.VariablesJsonEntry)
_sym_db.RegisterMessage(Trace.Error)
_sym_db.RegisterMessage(Trace.HTTP)
_sym_db.RegisterMessage(Trace.HTTP.Values)
_sym_db.RegisterMessage(Trace.HTTP.RequestHeadersEntry)
_sym_db.RegisterMessage(Trace.HTTP.ResponseHeadersEntry)
_sym_db.RegisterMessage(Trace.Location)
_sym_db.RegisterMessage(Trace.Node)
_sym_db.RegisterMessage(Trace.QueryPlanNode)
_sym_db.RegisterMessage(Trace.QueryPlanNode.SequenceNode)
_sym_db.RegisterMessage(Trace.QueryPlanNode.ParallelNode)
_sym_db.RegisterMessage(Trace.QueryPlanNode.FetchNode)
_sym_db.RegisterMessage(Trace.QueryPlanNode.FlattenNode)
_sym_db.RegisterMessage(Trace.QueryPlanNode.ResponsePathElement)

ReportHeader = _reflection.GeneratedProtocolMessageType(
    "ReportHeader",
    (_message.Message,),
    {
        "DESCRIPTOR": _REPORTHEADER,
        "__module__": "reports_pb2"
        # @@protoc_insertion_point(class_scope:ReportHeader)
    },
)
_sym_db.RegisterMessage(ReportHeader)

PathErrorStats = _reflection.GeneratedProtocolMessageType(
    "PathErrorStats",
    (_message.Message,),
    {
        "ChildrenEntry": _reflection.GeneratedProtocolMessageType(
            "ChildrenEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _PATHERRORSTATS_CHILDRENENTRY,
                "__module__": "reports_pb2"
                # @@protoc_insertion_point(class_scope:PathErrorStats.ChildrenEntry)
            },
        ),
        "DESCRIPTOR": _PATHERRORSTATS,
        "__module__": "reports_pb2"
        # @@protoc_insertion_point(class_scope:PathErrorStats)
    },
)
_sym_db.RegisterMessage(PathErrorStats)
_sym_db.RegisterMessage(PathErrorStats.ChildrenEntry)

QueryLatencyStats = _reflection.GeneratedProtocolMessageType(
    "QueryLatencyStats",
    (_message.Message,),
    {
        "DESCRIPTOR": _QUERYLATENCYSTATS,
        "__module__": "reports_pb2"
        # @@protoc_insertion_point(class_scope:QueryLatencyStats)
    },
)
_sym_db.RegisterMessage(QueryLatencyStats)

StatsContext = _reflection.GeneratedProtocolMessageType(
    "StatsContext",
    (_message.Message,),
    {
        "DESCRIPTOR": _STATSCONTEXT,
        "__module__": "reports_pb2"
        # @@protoc_insertion_point(class_scope:StatsContext)
    },
)
_sym_db.RegisterMessage(StatsContext)

ContextualizedQueryLatencyStats = _reflection.GeneratedProtocolMessageType(
    "ContextualizedQueryLatencyStats",
    (_message.Message,),
    {
        "DESCRIPTOR": _CONTEXTUALIZEDQUERYLATENCYSTATS,
        "__module__": "reports_pb2"
        # @@protoc_insertion_point(class_scope:ContextualizedQueryLatencyStats)
    },
)
_sym_db.RegisterMessage(ContextualizedQueryLatencyStats)

ContextualizedTypeStats = _reflection.GeneratedProtocolMessageType(
    "ContextualizedTypeStats",
    (_message.Message,),
    {
        "PerTypeStatEntry": _reflection.GeneratedProtocolMessageType(
            "PerTypeStatEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONTEXTUALIZEDTYPESTATS_PERTYPESTATENTRY,
                "__module__": "reports_pb2"
                # @@protoc_insertion_point(class_scope:ContextualizedTypeStats.PerTypeStatEntry)
            },
        ),
        "DESCRIPTOR": _CONTEXTUALIZEDTYPESTATS,
        "__module__": "reports_pb2"
        # @@protoc_insertion_point(class_scope:ContextualizedTypeStats)
    },
)
_sym_db.RegisterMessage(ContextualizedTypeStats)
_sym_db.RegisterMessage(ContextualizedTypeStats.PerTypeStatEntry)

FieldStat = _reflection.GeneratedProtocolMessageType(
    "FieldStat",
    (_message.Message,),
    {
        "DESCRIPTOR": _FIELDSTAT,
        "__module__": "reports_pb2"
        # @@protoc_insertion_point(class_scope:FieldStat)
    },
)
_sym_db.RegisterMessage(FieldStat)

TypeStat = _reflection.GeneratedProtocolMessageType(
    "TypeStat",
    (_message.Message,),
    {
        "PerFieldStatEntry": _reflection.GeneratedProtocolMessageType(
            "PerFieldStatEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _TYPESTAT_PERFIELDSTATENTRY,
                "__module__": "reports_pb2"
                # @@protoc_insertion_point(class_scope:TypeStat.PerFieldStatEntry)
            },
        ),
        "DESCRIPTOR": _TYPESTAT,
        "__module__": "reports_pb2"
        # @@protoc_insertion_point(class_scope:TypeStat)
    },
)
_sym_db.RegisterMessage(TypeStat)
_sym_db.RegisterMessage(TypeStat.PerFieldStatEntry)

ReferencedFieldsForType = _reflection.GeneratedProtocolMessageType(
    "ReferencedFieldsForType",
    (_message.Message,),
    {
        "DESCRIPTOR": _REFERENCEDFIELDSFORTYPE,
        "__module__": "reports_pb2"
        # @@protoc_insertion_point(class_scope:ReferencedFieldsForType)
    },
)
_sym_db.RegisterMessage(ReferencedFieldsForType)

Report = _reflection.GeneratedProtocolMessageType(
    "Report",
    (_message.Message,),
    {
        "TracesPerQueryEntry": _reflection.GeneratedProtocolMessageType(
            "TracesPerQueryEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _REPORT_TRACESPERQUERYENTRY,
                "__module__": "reports_pb2"
                # @@protoc_insertion_point(class_scope:Report.TracesPerQueryEntry)
            },
        ),
        "DESCRIPTOR": _REPORT,
        "__module__": "reports_pb2"
        # @@protoc_insertion_point(class_scope:Report)
    },
)
_sym_db.RegisterMessage(Report)
_sym_db.RegisterMessage(Report.TracesPerQueryEntry)

ContextualizedStats = _reflection.GeneratedProtocolMessageType(
    "ContextualizedStats",
    (_message.Message,),
    {
        "PerTypeStatEntry": _reflection.GeneratedProtocolMessageType(
            "PerTypeStatEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONTEXTUALIZEDSTATS_PERTYPESTATENTRY,
                "__module__": "reports_pb2"
                # @@protoc_insertion_point(class_scope:ContextualizedStats.PerTypeStatEntry)
            },
        ),
        "DESCRIPTOR": _CONTEXTUALIZEDSTATS,
        "__module__": "reports_pb2"
        # @@protoc_insertion_point(class_scope:ContextualizedStats)
    },
)
_sym_db.RegisterMessage(ContextualizedStats)
_sym_db.RegisterMessage(ContextualizedStats.PerTypeStatEntry)

TracesAndStats = _reflection.GeneratedProtocolMessageType(
    "TracesAndStats",
    (_message.Message,),
    {
        "ReferencedFieldsByTypeEntry": _reflection.GeneratedProtocolMessageType(
            "ReferencedFieldsByTypeEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _TRACESANDSTATS_REFERENCEDFIELDSBYTYPEENTRY,
                "__module__": "reports_pb2"
                # @@protoc_insertion_point(class_scope:TracesAndStats.ReferencedFieldsByTypeEntry)
            },
        ),
        "DESCRIPTOR": _TRACESANDSTATS,
        "__module__": "reports_pb2"
        # @@protoc_insertion_point(class_scope:TracesAndStats)
    },
)
_sym_db.RegisterMessage(TracesAndStats)
_sym_db.RegisterMessage(TracesAndStats.ReferencedFieldsByTypeEntry)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _TRACE_DETAILS_VARIABLESJSONENTRY._options = None
    _TRACE_DETAILS_VARIABLESJSONENTRY._serialized_options = b"8\001"
    _TRACE_HTTP_REQUESTHEADERSENTRY._options = None
    _TRACE_HTTP_REQUESTHEADERSENTRY._serialized_options = b"8\001"
    _TRACE_HTTP_RESPONSEHEADERSENTRY._options = None
    _TRACE_HTTP_RESPONSEHEADERSENTRY._serialized_options = b"8\001"
    _PATHERRORSTATS_CHILDRENENTRY._options = None
    _PATHERRORSTATS_CHILDRENENTRY._serialized_options = b"8\001"
    _CONTEXTUALIZEDTYPESTATS_PERTYPESTATENTRY._options = None
    _CONTEXTUALIZEDTYPESTATS_PERTYPESTATENTRY._serialized_options = b"8\001"
    _TYPESTAT_PERFIELDSTATENTRY._options = None
    _TYPESTAT_PERFIELDSTATENTRY._serialized_options = b"8\001"
    _REPORT_TRACESPERQUERYENTRY._options = None
    _REPORT_TRACESPERQUERYENTRY._serialized_options = b"8\001"
    _CONTEXTUALIZEDSTATS_PERTYPESTATENTRY._options = None
    _CONTEXTUALIZEDSTATS_PERTYPESTATENTRY._serialized_options = b"8\001"
    _TRACESANDSTATS_REFERENCEDFIELDSBYTYPEENTRY._options = None
    _TRACESANDSTATS_REFERENCEDFIELDSBYTYPEENTRY._serialized_options = b"8\001"
    _TRACE._serialized_start = 51
    _TRACE._serialized_end = 2648
    _TRACE_CACHEPOLICY._serialized_start = 660
    _TRACE_CACHEPOLICY._serialized_end = 781
    _TRACE_CACHEPOLICY_SCOPE._serialized_start = 736
    _TRACE_CACHEPOLICY_SCOPE._serialized_end = 781
    _TRACE_DETAILS._serialized_start = 784
    _TRACE_DETAILS._serialized_end = 930
    _TRACE_DETAILS_VARIABLESJSONENTRY._serialized_start = 878
    _TRACE_DETAILS_VARIABLESJSONENTRY._serialized_end = 930
    _TRACE_ERROR._serialized_start = 932
    _TRACE_ERROR._serialized_end = 1022
    _TRACE_HTTP._serialized_start = 1025
    _TRACE_HTTP._serialized_end = 1565
    _TRACE_HTTP_VALUES._serialized_start = 1270
    _TRACE_HTTP_VALUES._serialized_end = 1293
    _TRACE_HTTP_REQUESTHEADERSENTRY._serialized_start = 1295
    _TRACE_HTTP_REQUESTHEADERSENTRY._serialized_end = 1368
    _TRACE_HTTP_RESPONSEHEADERSENTRY._serialized_start = 1370
    _TRACE_HTTP_RESPONSEHEADERSENTRY._serialized_end = 1444
    _TRACE_HTTP_METHOD._serialized_start = 1446
    _TRACE_HTTP_METHOD._serialized_end = 1565
    _TRACE_LOCATION._serialized_start = 1567
    _TRACE_LOCATION._serialized_end = 1607
    _TRACE_NODE._serialized_start = 1610
    _TRACE_NODE._serialized_end = 1871
    _TRACE_QUERYPLANNODE._serialized_start = 1874
    _TRACE_QUERYPLANNODE._serialized_end = 2612
    _TRACE_QUERYPLANNODE_SEQUENCENODE._serialized_start = 2103
    _TRACE_QUERYPLANNODE_SEQUENCENODE._serialized_end = 2154
    _TRACE_QUERYPLANNODE_PARALLELNODE._serialized_start = 2156
    _TRACE_QUERYPLANNODE_PARALLELNODE._serialized_end = 2207
    _TRACE_QUERYPLANNODE_FETCHNODE._serialized_start = 2210
    _TRACE_QUERYPLANNODE_FETCHNODE._serialized_end = 2420
    _TRACE_QUERYPLANNODE_FLATTENNODE._serialized_start = 2422
    _TRACE_QUERYPLANNODE_FLATTENNODE._serialized_end = 2536
    _TRACE_QUERYPLANNODE_RESPONSEPATHELEMENT._serialized_start = 2538
    _TRACE_QUERYPLANNODE_RESPONSEPATHELEMENT._serialized_end = 2604
    _REPORTHEADER._serialized_start = 2651
    _REPORTHEADER._serialized_end = 2826
    _PATHERRORSTATS._serialized_start = 2829
    _PATHERRORSTATS._serialized_end = 3018
    _PATHERRORSTATS_CHILDRENENTRY._serialized_start = 2954
    _PATHERRORSTATS_CHILDRENENTRY._serialized_end = 3018
    _QUERYLATENCYSTATS._serialized_start = 3021
    _QUERYLATENCYSTATS._serialized_end = 3484
    _STATSCONTEXT._serialized_start = 3486
    _STATSCONTEXT._serialized_end = 3551
    _CONTEXTUALIZEDQUERYLATENCYSTATS._serialized_start = 3553
    _CONTEXTUALIZEDQUERYLATENCYSTATS._serialized_end = 3667
    _CONTEXTUALIZEDTYPESTATS._serialized_start = 3670
    _CONTEXTUALIZEDTYPESTATS._serialized_end = 3856
    _CONTEXTUALIZEDTYPESTATS_PERTYPESTATENTRY._serialized_start = 3795
    _CONTEXTUALIZEDTYPESTATS_PERTYPESTATENTRY._serialized_end = 3856
    _FIELDSTAT._serialized_start = 3859
    _FIELDSTAT._serialized_end = 4065
    _TYPESTAT._serialized_start = 4068
    _TYPESTAT._serialized_end = 4208
    _TYPESTAT_PERFIELDSTATENTRY._serialized_start = 4133
    _TYPESTAT_PERFIELDSTATENTRY._serialized_end = 4196
    _REFERENCEDFIELDSFORTYPE._serialized_start = 4210
    _REFERENCEDFIELDSFORTYPE._serialized_end = 4278
    _REPORT._serialized_start = 4281
    _REPORT._serialized_end = 4493
    _REPORT_TRACESPERQUERYENTRY._serialized_start = 4423
    _REPORT_TRACESPERQUERYENTRY._serialized_end = 4493
    _CONTEXTUALIZEDSTATS._serialized_start = 4496
    _CONTEXTUALIZEDSTATS._serialized_end = 4723
    _CONTEXTUALIZEDSTATS_PERTYPESTATENTRY._serialized_start = 3795
    _CONTEXTUALIZEDSTATS_PERTYPESTATENTRY._serialized_end = 3856
    _TRACESANDSTATS._serialized_start = 4726
    _TRACESANDSTATS._serialized_end = 5039
    _TRACESANDSTATS_REFERENCEDFIELDSBYTYPEENTRY._serialized_start = 4952
    _TRACESANDSTATS_REFERENCEDFIELDSBYTYPEENTRY._serialized_end = 5039
# @@protoc_insertion_point(module_scope)



from typing import Callable, Dict, Optional, Sequence, TYPE_CHECKING, Type

from graphene import Argument, Enum, List
from sqlalchemy.orm import ColumnProperty
from sqlalchemy.sql.elements import UnaryExpression
from sqlalchemy.types import Enum as SQLAlchemyEnumType

from .registry import Registry
from .utils import get_object_type_manual_fields, to_enum_value_name, to_type_name

if TYPE_CHECKING:
    from .types import SQLAlchemyObjectType


def _convert_sa_to_graphene_enum(sa_enum: SQLAlchemyEnumType, fallback_name=None):
    """Convert the given SQLAlchemy Enum type to a Graphene Enum type.

    The name of the Graphene Enum will be determined as follows:
    If the SQLAlchemy Enum is based on a Python Enum, use the name
    of the Python Enum.  Otherwise, if the SQLAlchemy Enum is named,
    use the SQL name after conversion to a type name. Otherwise, use
    the given fallback_name or raise an error if it is empty.

    The Enum value names are converted to upper case if necessary.
    """
    if not isinstance(sa_enum, SQLAlchemyEnumType):
        raise TypeError(f"Expected sqlalchemy.types.Enum, but got: {sa_enum!r}")
    enum_class = sa_enum.enum_class
    if enum_class:
        if all(to_enum_value_name(key) == key for key in enum_class.__members__):
            return Enum.from_enum(enum_class)
        name = enum_class.__name__
        members = [
            (to_enum_value_name(key), value.value)
            for key, value in enum_class.__members__.items()
        ]
    else:
        sql_enum_name = sa_enum.name
        if sql_enum_name:
            name = to_type_name(sql_enum_name)
        elif fallback_name:
            name = fallback_name
        else:
            raise TypeError(f"No type name specified for {sa_enum!r}")
        members = [(to_enum_value_name(key), key) for key in sa_enum.enums]
    return Enum(name, members)


def enum_for_sa_enum(sa_enum: SQLAlchemyEnumType, registry: Registry):
    """Return the Graphene Enum type for the specified SQLAlchemy Enum type."""
    if not isinstance(sa_enum, SQLAlchemyEnumType):
        raise TypeError(f"Expected sqlalchemy.types.Enum, but got: {sa_enum!r}")
    enum = registry.get_graphene_enum_for_sa_enum(sa_enum)
    if not enum:
        enum = _convert_sa_to_graphene_enum(sa_enum)
        registry.register_enum(sa_enum, enum)
    return enum


def enum_for_field(obj_type: Type["SQLAlchemyObjectType"], field_name: str):
    """Return the Graphene Enum type for the specified Graphene field."""
    from .types import SQLAlchemyObjectType

    if not isinstance(obj_type, type) or not issubclass(obj_type, SQLAlchemyObjectType):
        raise TypeError(f"Expected SQLAlchemyObjectType, but got: {obj_type!r}")
    if not field_name or not isinstance(field_name, str):
        raise TypeError(f"Expected a field name, but got: {field_name!r}")
    registry = obj_type._meta.registry
    orm_field = registry.get_orm_field_for_graphene_field(obj_type, field_name)
    if orm_field is None:
        raise TypeError(f"Cannot get {obj_type._meta.name}.{field_name}")
    if not isinstance(orm_field, ColumnProperty):
        raise TypeError(
            f"{obj_type._meta.name}.{field_name} does not map to model column"
        )
    column = orm_field.columns[0]
    sa_enum = column.type
    if not isinstance(sa_enum, SQLAlchemyEnumType):
        raise TypeError(
            f"{obj_type._meta.name}.{field_name} does not map to enum column"
        )
    enum = registry.get_graphene_enum_for_sa_enum(sa_enum)
    if not enum:
        fallback_name = obj_type._meta.name + to_type_name(field_name)
        enum = _convert_sa_to_graphene_enum(sa_enum, fallback_name)
        registry.register_enum(sa_enum, enum)
    return enum


def _default_sort_enum_symbol_name(column_name: str, sort_asc=True):
    return to_enum_value_name(column_name) + ("_ASC" if sort_asc else "_DESC")


def sort_enum_for_object_type(
    obj_type: Type["SQLAlchemyObjectType"],
    name: Optional[str] = None,
    only_fields: Optional[Sequence[str]] = None,
    only_indexed: Optional[bool] = None,
    get_symbol_name: Optional[Callable] = None,
    extra_members: Optional[Dict[str, UnaryExpression]] = None,
):
    """Return Graphene Enum for sorting the given SQLAlchemyObjectType.

    Parameters
    - obj_type: The object type for which the sort Enum shall be generated.
    - name: Name to use for the sort Enum.
        If not provided, it will be set to the object type name + 'SortEnum'
    - only_fields: If this is set, only fields from this sequence will be considered.
    - only_indexed : If this is set, only indexed columns will be considered.
    - get_symbol_name : Function which takes the column name and a boolean indicating
        if the sort direction is ascending, and returns the symbol name
        for the current column and sort direction. If no such function
        is passed, a default function will be used that creates the symbols
        'foo_asc' and 'foo_desc' for a column with the name 'foo'.

    Returns
    - Enum
        The Graphene Enum type
    """
    name = name or obj_type._meta.name + "SortEnum"
    registry = obj_type._meta.registry
    enum = registry.get_sort_enum_for_object_type(obj_type)
    custom_options = dict(
        only_fields=only_fields,
        only_indexed=only_indexed,
        get_symbol_name=get_symbol_name,
    )
    if enum:
        if name != enum.__name__ or custom_options != enum.custom_options:
            raise ValueError(f"Sort enum for {obj_type} has already been customized")
    else:
        members = {}
        default = []
        fields = obj_type._meta.fields
        get_name = get_symbol_name or _default_sort_enum_symbol_name
        for field_name in fields:
            if only_fields and field_name not in only_fields:
                continue
            orm_field = registry.get_orm_field_for_graphene_field(obj_type, field_name)
            if not isinstance(orm_field, ColumnProperty):
                continue
            column = orm_field.columns[0]
            if only_indexed and not (column.primary_key or column.index):
                continue
            asc_name = get_name(field_name, True)
            asc_value = column.asc()
            desc_name = get_name(field_name, False)
            desc_value = column.desc()
            if column.primary_key:
                default.append(asc_value)
            members.update({asc_name: asc_value, desc_name: desc_value})

        extra_fields = get_object_type_manual_fields(obj_type)
        for field_name, attr in extra_fields.items():
            if hasattr(attr, "model_field"):
                column = getattr(attr, "model_field", None)
                if column is not None:
                    asc_name = get_name(field_name, True)
                    asc_value = column.asc()
                    desc_name = get_name(field_name, False)
                    desc_value = column.desc()
                    if column.primary_key:
                        default.append(asc_value)
                    members.update({asc_name: asc_value, desc_name: desc_value})

        if extra_members:
            members.update(extra_members)

        enum = Enum(name, members.items())
        enum.default = default  # store default as attribute
        enum.custom_options = custom_options
        registry.register_sort_enum(obj_type, enum)
    return enum


def sort_argument_for_object_type(
    obj_type: Type["SQLAlchemyObjectType"],
    enum_name: Optional[str] = None,
    only_fields: Optional[Sequence] = None,
    only_indexed: Optional[bool] = None,
    get_symbol_name: Optional[Callable] = None,
    has_default: bool = True,
):
    """Returns Graphene Argument for sorting the given SQLAlchemyObjectType.

    Parameters
    - obj_type: The object type for which the sort Argument shall be generated.
    - enum_name: Name to use for the sort Enum.
        If not provided, it will be set to the object type name + 'SortEnum'
    - only_fields: If this is set, only fields from this sequence will be considered.
    - only_indexed: If this is set, only indexed columns will be considered.
    - get_symbol_name: Function which takes the column name and a boolean indicating
        if the sort direction is ascending, and returns the symbol name
        for the current column and sort direction. If no such function
        is passed, a default function will be used that creates the symbols
        'foo_asc' and 'foo_desc' for a column with the name 'foo'.
    - has_default: If this is set to False, no sorting will happen when this argument is not
        passed. Otherwise results will be sortied by the primary key(s) of the model.

    Returns
    - Enum
        A Graphene Argument that accepts a list of sorting directions for the model.
    """
    enum = obj_type.sort_enum(
        name=enum_name,
        only_fields=only_fields,
        only_indexed=only_indexed,
        get_symbol_name=get_symbol_name,
    )
    if not has_default:
        enum.default = None

    return Argument(List(enum), default_value=enum.default)



import base64
import json
from typing import NamedTuple, Tuple, Type, TypeVar, Union

_T = TypeVar("_T")


def decode_gql_id(encoded_id: str) -> Tuple[str, Union[str, int]]:
    try:
        text = base64.b64decode(encoded_id).decode()
    except Exception as e:
        raise ValueError("Invalid base64")

    type_name, id_text = text.split(":")
    return type_name, json.loads(id_text)


def encode_gql_id(type_: str, id_: Union[str, int]) -> str:
    text = f"{type_}:{json.dumps(id_)}"
    return base64.b64encode(text.encode()).decode()


class ResolvedGlobalId(NamedTuple):
    type: str
    id: Union[str, int]

    def encode(self) -> str:
        return encode_gql_id(*self)

    @classmethod
    def decode(cls: Type[_T], global_id: str) -> _T:
        return cls(*decode_gql_id(global_id))



import logging
from typing import Optional, TYPE_CHECKING, Type

import sqlalchemy as sa
from graphene import Connection, PageInfo
from graphene.types import ResolveInfo
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import DeclarativeMeta

from .utils import (
    get_offset_with_default,
    offset_to_cursor,
)
from ..query_helper import QueryHelper
from ..utils import filter_requested_fields_for_object

if TYPE_CHECKING:
    from ..fields import UnsortedSQLAlchemyConnectionField
DEFAULT_LIMIT = 1000


def get_count_query(query, model):
    only_q = query.with_only_columns(
        *sa.inspect(model).primary_key,
    ).order_by(None)
    return sa.select([sa.func.count()]).select_from(only_q.alias())


def construct_page_info(
    cls: Type[PageInfo],
    info: ResolveInfo,
    edges: list,
    limit: int,
    offset: int,
) -> PageInfo:
    page_info_kwargs = {}

    page_info_fields = QueryHelper.get_page_info_fields(info)

    if not page_info_fields:
        return cls()

    for field in page_info_fields:
        if field == "has_previous_page":
            page_info_kwargs[field] = offset > 0
        elif field == "has_next_page":
            page_info_kwargs[field] = limit and len(edges) > limit
        elif edges:
            if field == "start_cursor":
                page_info_kwargs[field] = edges[0].cursor
            elif field == "end_cursor":
                page_info_kwargs[field] = edges[:limit][-1].cursor

    return cls(**page_info_kwargs)


async def connection_from_query(
    cls: Type["UnsortedSQLAlchemyConnectionField"],
    model: Type[DeclarativeMeta],
    info: ResolveInfo,
    args: Optional[dict] = None,
    connection_type: Type[Connection] = Connection,
) -> Connection:
    """
    Given a slice (subset) of an array, returns a connection object for use in
    GraphQL.
    This function is similar to `connectionFromArray`, but is intended for use
    cases where you know the cardinality of the connection, consider it too large
    to materialize the entire array, and instead wish pass in a slice of the
    total result large enough to cover the range specified in `args`.
    """
    args = args or {}
    session: AsyncSession = info.context.session

    # has_last_arg = QueryHelper.has_arg(info, "last")
    # if not QueryHelper.get_filters(info) and has_last_arg:
    #     raise TypeError('Cannot set "last" without filters applied')

    edge_type = connection_type.Edge
    node_type = edge_type.node.type
    page_info_type = getattr(connection_type, "PageInfo", PageInfo)

    before = args.get("before")
    after = args.get("after")
    first = args.get("first")
    last = args.get("last")

    current_field = QueryHelper.get_current_field(info)
    has_total_count = current_field and "total_count" in {
        i.name for i in current_field.values
    }

    if (before, after, first, last) == (None, None, None, None):
        first = DEFAULT_LIMIT
        logging.warning(f"Query without border, {first=}")

    query = await cls.get_query(
        model=model,
        info=info,
        cls_name=node_type.__name__,
        **args,
    )

    total_count = None
    # TODO: Move total_count to PageInfo
    if (last and not before) or (after and not first and not before) or has_total_count:
        count_query = get_count_query(query, model)
        right_offset = total_count = (await session.execute(count_query)).scalar()
    else:
        right_offset = get_offset_with_default(before)

    left_offset = get_offset_with_default(after) + 1 if after else 0

    if isinstance(first, int):
        right_offset = min(left_offset + first, right_offset)
    if isinstance(last, int):
        left_offset = max(left_offset, right_offset - last)

    _slice = query
    # If supplied slice is too large, trim it down before mapping over it.
    limit = first or last or (right_offset - left_offset)

    if limit:
        _slice = _slice.limit(limit + 1)
    if left_offset:
        _slice = _slice.offset(left_offset)

    edges = []

    for i, v in enumerate(await session.execute(_slice)):
        node_value = filter_requested_fields_for_object(dict(v), node_type)
        edge = edge_type(
            node=node_type(**node_value),
            cursor=offset_to_cursor(left_offset + i),
        )
        edges.append(edge)

    connection = connection_type(
        edges=edges[:limit],
        page_info=construct_page_info(
            cls=page_info_type,
            info=info,
            edges=edges,
            limit=limit,
            offset=left_offset,
        ),
    )

    if has_total_count:
        connection.total_count = total_count

    return connection
